from herre.wards.graphql import ParsedQuery


GET_TABLE = ParsedQuery(
    """
query Table($id: ID!){
  table(id: $id){
    id
    name
    tags
    store
    creator {
      email
    }
    sample {
      id
    }
    representation {
      id
    }
    experiment {
      id
    }
  
  }
}
"""
)



from herre.wards.graphql import ParsedQuery



CREATE_SAMPLE = ParsedQuery("""
mutation SampleCreate($name: String, $creator: String, $meta: GenericScalar, $experiments: [ID]) {
  createSample(name: $name, creator: $creator, meta: $meta, experiments: $experiments){
    id
    name
    creator {
        email
    }
  }
}
""")


from herre.wards.graphql import ParsedQuery


CREATE_METRIC = ParsedQuery("""
mutation CreateMetric($rep:ID!, $key: String!, $value: GenericScalar!){
  createMetric(rep: $rep, key: $key, value: $value){
    id
    rep {
      id
    }
    key
    value
    creator {
      id
    }
    createdAt
  }
  
}
""")


from herre.wards.graphql import ParsedQuery


CREATE_THUMBNAIL = ParsedQuery(
    """
mutation CreateThumbnail($file: ImageFile!, $rep: ID!) {
  uploadThumbnail(file: $file, rep: $rep){
    id
    representation {
      id
    }
    image
  }
}
"""
)



from herre.wards.graphql import ParsedQuery





CREATE_EXPERIMENT = ParsedQuery("""
mutation CreateExperiment($name: String!, $creator: String, $meta: GenericScalar, $description: String) {
  createExperiment(name: $name, creator: $creator, description: $description, meta: $meta){
    id
    name
    creator {
        email
    }
    meta
  }
}
""")


from herre.wards.graphql import ParsedQuery



GET_METRIC = ParsedQuery("""
query GetMetric($id: ID!){
  metric(id: $id){
    id
    rep {
      id
    }
    key
    value
    creator {
      id
    }
    createdAt
  }
}
""")


from fakts import config
from mikro.ward import MikroWard
from herre.wards.registry import get_ward_registry
import os
import s3fs
import xarray as xr


class RepresentationException(Exception):
    pass


class Array:
    def _getZarrStore(self):
        ward: MikroWard = get_ward_registry().get_ward_instance("mikro")
        transcript = ward.transcript
        protocol = "https" if ward.config.s3.secure else "http"
        endpoint_url = f"{protocol}://{ward.config.s3.host}:{ward.config.s3.port}"

        os.environ["AWS_ACCESS_KEY_ID"] = transcript.params.access_key
        os.environ["AWS_SECRET_ACCESS_KEY"] = transcript.params.secret_key

        s3_path = f"zarr/{self.store}"
        store = s3fs.S3FileSystem(client_kwargs={"endpoint_url": endpoint_url})
        return store.get_mapper(s3_path)

    @property
    def data(self) -> xr.DataArray:
        assert (
            self.store is not None
        ), "Please query 'store' in your request on 'Representation'. Data is not accessible otherwise"
        return xr.open_zarr(store=self._getZarrStore(), consolidated=True)["data"]

    def save_array(self, array: xr.DataArray, compute=True, chunks=None):
        apiversion = "0.1"
        fileversion = "0.1"

        if "x" not in array.dims:
            raise RepresentationException(
                "Representations must always have a 'x' Dimension"
            )

        if "y" not in array.dims:
            raise RepresentationException(
                "Representations must always have a 'y' Dimension"
            )

        if "t" not in array.dims:
            array = array.expand_dims("t")
        if "c" not in array.dims:
            array = array.expand_dims("c")
        if "z" not in array.dims:
            array = array.expand_dims("z")

        chunks = chunks or {
            "t": 1,
            "x": array.sizes["x"],
            "y": array.sizes["y"],
            "z": 1,
        }

        array = array.chunk(
            {key: chunksize for key, chunksize in chunks.items() if key in array.dims}
        )
        if apiversion == "0.1":
            dataset = array.to_dataset(name="data")
            dataset.attrs["apiversion"] = apiversion
            dataset.attrs["fileversion"] = fileversion
            if fileversion == "0.1":
                dataset.attrs["model"] = str(self.Meta.identifier)
                dataset.attrs["unique"] = str(self.unique)
            else:
                raise NotImplementedError(
                    "This FileVersion has not been Implemented yet"
                )
        else:
            raise NotImplementedError("This API Version has not been Implemented Yet")

        return dataset.to_zarr(
            store=self._getZarrStore(), consolidated=True, compute=compute
        )



from arkitekt.packers.structure import BoundType
from herre.access.object import GraphQLObject
from mikro.graphql.mutations.experiment import CREATE_EXPERIMENT
from mikro.graphql.mutations.metric import CREATE_METRIC
from mikro.graphql.mutations.omerofile import CREATE_OMERO_FILE
from mikro.graphql.mutations.representation import (
    CREATE_REPRESENTATION,
    UPDATE_REPRESENTATION,
)
from mikro.graphql.mutations.table import CREATE_TABLE, UPDATE_TABLE
from mikro.graphql.mutations.thumbnail import CREATE_THUMBNAIL
from mikro.graphql.queries.omerofile import GET_OMEROFILE
from mikro.graphql.queries.sample import GET_SAMPLE
from mikro.graphql.mutations.sample import CREATE_SAMPLE
from mikro.array import Array
from mikro.graphql.queries.table import GET_TABLE
from mikro.manager import (
    AsyncRepresentationManager,
    AsyncTableManager,
    SyncRepresentationManager,
    SyncTableManager,
)
from typing import Any, List, Optional
from herre.convenience import GraphQLModel
import xarray as xr
from typing import ForwardRef
from enum import Enum
from mikro.parquet import Parquet
from mikro.ward import MikroWard
from mikro.graphql.queries.experiment import GET_EXPERIMENT
from mikro.graphql.queries.representation import GET_REPRESENTATION
from mikro.widgets import MY_TOP_REPRESENTATIONS, MY_TOP_SAMPLES

Representation = ForwardRef("Representation")
Table = ForwardRef("Table")
Sample = ForwardRef("Sample")
Experiment = ForwardRef("Experiment")


class RepresentationVariety(str, Enum):
    VOXEL = "VOXEL"
    MASK = "MASK"
    UNKNOWN = "UNKNOWN"


class Metric(GraphQLModel):
    key: Optional[str]
    value: Optional[Any]

    class Meta:
        identifier = "metric"
        ward = "mikro"
        create = CREATE_METRIC


class Sample(GraphQLModel):
    name: Optional[str]
    meta: Optional[dict]
    representations: Optional[List[Representation]]
    experiments: Optional[List[Experiment]]
    name: Optional[str]

    class Meta:
        identifier = "sample"
        ward = "mikro"
        create = CREATE_SAMPLE
        get = GET_SAMPLE
        widget = MY_TOP_SAMPLES


class Experiment(GraphQLModel):
    meta: Optional[dict]
    id: Optional[int]
    name: Optional[str]
    description: Optional[str]
    descriptionLong: Optional[str]
    samples: Optional[List[Sample]]

    class Meta:
        identifier = "experiment"
        ward = "mikro"
        get = GET_EXPERIMENT
        create = CREATE_EXPERIMENT


class Plane(GraphQLObject):
    zIndex: Optional[int] = 0
    yIndex: Optional[int] = 0
    xIndex: Optional[int] = 0
    cIndex: Optional[int] = 0
    tIndex: Optional[int] = 0
    exposureTime: Optional[float] = 0
    deltaT: Optional[float] = 0


class PhysicalSize(GraphQLObject):
    x: Optional[float] = 1
    y: Optional[float] = 1
    z: Optional[float] = 1
    t: Optional[float] = 1


class Channel(GraphQLObject):
    name: Optional[str] = "Test"
    emmissionWavelength: Optional[float] = 0
    excitationWavelength: Optional[float] = 0
    acquisitionMode: Optional[str] = "Standard"
    color: Optional[str] = "rgb(244,255,232)"


class OmeroRepresentation(GraphQLObject):
    physicalSize: Optional[PhysicalSize]
    channels: Optional[List[Channel]]
    planes: Optional[List[Plane]]


class OmeroFileType(str, Enum):
    MSR = "MSR"
    TIFF = "TIFF"
    UNKNOWN = "UNKNOWN"


class OmeroFile(GraphQLModel):
    file: Optional[str]
    name: Optional[str]
    type: Optional[OmeroFileType]

    class Meta:
        identifier = "omerofile"
        ward = "mikro"
        create = CREATE_OMERO_FILE
        get = GET_OMEROFILE


class Representation(GraphQLModel, Array):
    """Representation

    A Representation is a Five Dimensional Array that lets you access data

    Args:
        GraphQLStructure ([type]): [description]
    """

    omero: Optional[OmeroRepresentation]
    meta: Optional[dict]
    name: Optional[str]
    package: Optional[str]
    store: Optional[str]
    shape: Optional[List[int]]
    image: Optional[str]
    unique: Optional[str]
    variety: Optional[RepresentationVariety]
    sample: Optional[Sample]
    tags: Optional[List[str]]
    tables: Optional[List[Table]]
    metrics: Optional[List[Metric]]
    thumbnail: Optional[str]
    origins: Optional[List[Representation]]
    derived: Optional[List[Representation]]

    asyncs = AsyncRepresentationManager()
    objects = SyncRepresentationManager()

    class Meta:
        identifier = "representation"
        ward = "mikro"
        get = GET_REPRESENTATION
        create = CREATE_REPRESENTATION
        update = UPDATE_REPRESENTATION
        widget = MY_TOP_REPRESENTATIONS


class Render(GraphQLModel):
    representation: Optional[Representation]

    class Meta:
        identifier = "render"
        ward = "mikro"


class Thumbnail(GraphQLModel):
    representation: Optional[Representation]
    image: Optional[str]

    class Meta:
        identifier = "thumbnail"
        ward = "mikro"
        create = CREATE_THUMBNAIL


class Table(GraphQLModel, Parquet):
    id: Optional[int]
    name: Optional[str]
    store: Optional[str]
    columns: Optional[List[str]]
    query: Optional[List[List[Any]]]

    asyncs = AsyncTableManager()
    objects = SyncTableManager()

    class Meta:
        identifier = "table"
        ward = "mikro"
        bound = BoundType.APP
        create = CREATE_TABLE
        update = UPDATE_TABLE
        get = GET_TABLE


Representation.update_forward_refs()
Table.update_forward_refs()
Sample.update_forward_refs()
Experiment.update_forward_refs()



from herre.wards.graphql import ParsedQuery


CREATE_TABLE = ParsedQuery(
    """
mutation CreateTable($sample: ID, $name: String, $tags: [String], $creator: String, $representation: ID, $experiment: ID){
  createTable(sample: $sample, name: $name, representation: $representation, experiment: $experiment, tags: $tags, creator: $creator){
    name
    id
    tags
    store
  }
}
"""
)

UPDATE_TABLE = ParsedQuery(
    """
mutation UpdateTable($id: ID!){
  updateTable(id: $id){
    name
    id
    tags
    store
  }
}
"""
)



from concurrent.futures.thread import ThreadPoolExecutor
from herre.access.model.graphql import GraphQLSyncModelManager
from herre.access.model import GraphQLAsyncModelManager
from herre.excecutors import default_executor
import asyncio
import xarray as xr
import pandas as pd


class AsyncRepresentationManager(GraphQLAsyncModelManager):
    async def from_xarray(self, array: xr.DataArray, compute=True, **kwargs):
        instance = await self.create(**kwargs)
        executor = default_executor.get()
        if executor:
            co_future = executor.submit(instance.save_array, array, compute=True)
            await asyncio.wrap_future(co_future)
        else:
            with ThreadPoolExecutor(max_workers=4) as executor:
                co_future = executor.submit(instance.save_array, array, compute=True)
                await asyncio.wrap_future(co_future)
        instance = await self.update(id=instance.id, **kwargs)
        return instance

    async def update(self, **kwargs) -> "Representation":
        return await self.from_query(self.modelClass.Meta.update, **kwargs)


class SyncRepresentationManager(GraphQLSyncModelManager):
    def from_xarray(self, array: xr.DataArray, compute=True, **kwargs):
        instance = self.create(**kwargs)
        instance.save_array(array, compute=compute)
        instance = self.update(id=instance.id, **kwargs)
        return instance

    def update(self, **kwargs) -> "Representation":
        return self.from_query(self.modelClass.Meta.update, **kwargs)


class AsyncTableManager(GraphQLAsyncModelManager):
    async def update(self, **kwargs) -> "Representation":
        return await self.from_query(self.modelClass.Meta.update, **kwargs)


class SyncTableManager(GraphQLSyncModelManager):
    def from_df(self, df: pd.DataFrame, **kwargs) -> "Table":
        instance = self.create(**kwargs)
        instance.save_df(df)
        instance = self.update(id=instance.id, **kwargs)
        return instance

    def update(self, **kwargs) -> "Table":
        return self.from_query(self.modelClass.Meta.update, **kwargs)



from herre.wards.graphql import ParsedQuery


FILTER_REPRESENTATION = ParsedQuery(
    """
query Representation($name: String) {
  representations(name: $name) {
    id
    name
    store
    variety
    tags
    unique
    creator {
      email
    }
    sample {
      id
    }
    meta
  }
}
"""
)

GET_REPRESENTATION = ParsedQuery(
    """
query Representation($id: ID!){
  representation(id: $id){
    id
    name
    tags
    variety
    store
    meta
    unique
    creator {
      email
    }
    sample {
      id
    }
  }
}
"""
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mikro', 'mikro.graphql', 'mikro.graphql.mutations', 'mikro.graphql.queries']

package_data = \
{'': ['*']}

install_requires = \
['dask==2021.11.2',
 'docstring-parser>=0.10,<0.11',
 'herre>=0.1.58,<0.2.0',
 'inflection>=0.5.1,<0.6.0',
 'pandas>=1.3.4,<2.0.0',
 'pyarrow>=6.0.1,<7.0.0',
 's3fs>=2021.9.0,<2022.0.0',
 'xarray>=0.19.0,<0.20.0',
 'zarr==2.8.3']

setup_kwargs = {
    'name': 'mikro',
    'version': '0.1.66',
    'description': 'images for arnheim',
    'long_description': None,
    'author': 'jhnnsrs',
    'author_email': 'jhnnsrs@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from pydantic.main import BaseModel
from fakts import Fakts, get_current_fakts, Config
from herre.herre import Herre
from herre.wards.query import TypedQuery
from herre.access.object import GraphQLObject
from herre import get_current_herre
from herre.wards.graphql import ParsedQuery, GraphQLWard
import aiohttp
import xarray as xr


class S3Config(BaseModel):
    host: str
    port: int
    secure: bool


class MikroConfig(Config):
    host: str
    port: int
    secure: bool
    s3: S3Config

    class Config:
        group = "mikro"
        env_prefix = "mikro_"

    @property
    def endpoint(self):
        return f"http://{self.host}:{self.port}/graphql"


class AccessParams(GraphQLObject):
    access_key: str
    secret_key: str


class Transcript(GraphQLObject):
    protocol: str
    path: str
    params: AccessParams


class MikroWard(GraphQLWard):
    configClass = MikroConfig
    config: MikroConfig

    class Meta:
        key = "mikro"

    async def negotiate(self):
        transcript_query = await self.arun(
            ParsedQuery(
                """mutation Negotiate {
            negotiate 
        }"""
            )
        )
        return Transcript(**transcript_query["negotiate"])


class playground:
    def __init__(self, width=900, height=700) -> None:
        herre = get_current_herre()
        self.width = width
        self.height = height

    def _repr_html_(self):
        return f"<iframe src='http://{self.config.host}:{self.config.port}/graphql' width={self.width} height={self.height}></iframe>"


class gql(TypedQuery):
    ward_key = "mikro"



from herre.wards.graphql import ParsedQuery


CREATE_OMERO_FILE = ParsedQuery(
    """
mutation UploadOmeroFile($file: Upload!){
  uploadOmeroFile(file:$file) {
    id
    file
    type
    name
  }
}
"""
)



try:
    from arkitekt.schema.widgets import SliderWidget, SearchWidget


    MY_TOP_REPRESENTATIONS = SearchWidget(query="""
                    query Search($search: String){
                        options: representations(name: $search){
                            value: id
                            label: name
                        }
                    }
                    """)


    MY_TOP_SAMPLES = SearchWidget(query="""
                    query Search($search: String){
                        options: samples(name: $search){
                            value: id
                            label: name
                        }
                    }
                    """)

except:
    MY_TOP_REPRESENTATIONS = None
    MY_TOP_SAMPLES = None


from mikro.ward import MikroWard
from herre.wards.registry import get_ward_registry
import os
import s3fs
import xarray as xr
import pyarrow.parquet as pq
from pyarrow import Table as ArrowTable
import pandas as pd


class Parquet:
    def _getFileSystem(self):
        ward: MikroWard = get_ward_registry().get_ward_instance("mikro")
        transcript = ward.transcript
        protocol = "https" if ward.config.s3.secure else "http"
        endpoint_url = f"{protocol}://{ward.config.s3.host}:{ward.config.s3.port}"

        os.environ["AWS_ACCESS_KEY_ID"] = transcript.params.access_key
        os.environ["AWS_SECRET_ACCESS_KEY"] = transcript.params.secret_key

        s3_path = f"parquet/{self.store}"
        s3 = s3fs.S3FileSystem(client_kwargs={"endpoint_url": endpoint_url})
        return s3

    @property
    def data(self) -> pd.DataFrame:
        assert (
            self.store is not None
        ), "Please query 'store' in your request on 'Table'. Data is not accessible otherwise"
        s3_path = f"parquet/{self.store}"
        return (
            pq.ParquetDataset(s3_path, filesystem=self._getFileSystem())
            .read_pandas()
            .to_pandas()
        )

    def save_df(self, df: pd.DataFrame):
        table: ArrowTable = ArrowTable.from_pandas(df)
        s3_path = f"parquet/{self.store}"
        pq.write_table(table, s3_path, filesystem=self._getFileSystem())



from herre.wards.graphql import ParsedQuery

REP_FRAGMENT = """
  name
  id
  variety
  tags
  store
  unique
  meta
  origins {
    id
    name
  }
"""


CREATE_REPRESENTATION = ParsedQuery(
    """
mutation Representation($sample: ID, $name: String, $tags: [String], $variety: RepresentationVarietyInput, $creator: String, $meta: GenericScalar, $omero: OmeroRepresentationInput, $origins: [ID]){
  createRepresentation(sample: $sample, name: $name, tags: $tags, variety: $variety, creator: $creator, meta: $meta, omero: $omero, origins: $origins){
    """
    + REP_FRAGMENT
    + """
  
  }
}
"""
)

UPDATE_REPRESENTATION = ParsedQuery(
    """
mutation Representation($id: ID!){
  updateRepresentation(rep: $id){
    """
    + REP_FRAGMENT
    + """

  }
}
"""
)
